#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#	tcprelay.py - TCP connection relay for usbmuxd
#
# Original work Copyright (C) 2009  Hector Martin "marcan" <hector@marcansoft.com>
# Modified work Copyright (C) 2019  phx <https://github.com/phx>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 or version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import select
import sys
import usbmux
from optparse import OptionParser

if sys.version_info > (3, 0):
    python3 = True
else:
    python3 = False

if python3:
    import socketserver
else:
    import SocketServer as socketserver


class SocketRelay(object):
    def __init__(self, a, b, maxbuf=65535):
        self.a = a
        self.b = b
        if python3:
            self.atob = b''
            self.btoa = b''
        else:
            self.atob = ''
            self.btoa = ''
        self.maxbuf = maxbuf

    def handle(self):
        while True:
            rlist = []
            wlist = []
            xlist = [self.a, self.b]
            if self.atob:
                wlist.append(self.b)
            if self.btoa:
                wlist.append(self.a)
            if len(self.atob) < self.maxbuf:
                rlist.append(self.a)
            if len(self.btoa) < self.maxbuf:
                rlist.append(self.b)
            rlo, wlo, xlo = select.select(rlist, wlist, xlist)
            if xlo:
                return
            if self.a in wlo:
                n = self.a.send(self.btoa)
                self.btoa = self.btoa[n:]
            if self.b in wlo:
                n = self.b.send(self.atob)
                self.atob = self.atob[n:]
            if self.a in rlo:
                s = self.a.recv(self.maxbuf - len(self.atob))
                if not s:
                    return
                self.atob += s
            if self.b in rlo:
                s = self.b.recv(self.maxbuf - len(self.btoa))
                if not s:
                    return
                self.btoa += s


class TCPRelay(socketserver.BaseRequestHandler):
    def handle(self):
        print("Incoming connection to %d" % self.server.server_address[1])
        mux = usbmux.USBMux(options.sockpath)
        print("Waiting for devices...")
        if not mux.devices:
            mux.process(1.0)
        if not mux.devices:
            print("No device found")
            self.request.close()
            return
        dev = mux.devices[0]
        print("Connecting to device %s" % str(dev))
        dsock = mux.connect(dev, self.server.rport)
        lsock = self.request
        print("Connection established, relaying data")
        try:
            fwd = SocketRelay(dsock, lsock, self.server.bufsize * 1024)
            fwd.handle()
        finally:
            dsock.close()
            lsock.close()
        print("Connection closed")


class TCPServer(socketserver.TCPServer):
    allow_reuse_address = True


class ThreadedTCPServer(socketserver.ThreadingMixIn, TCPServer):
    pass


HOST = "localhost"

parser = OptionParser(
    usage="usage: %prog [OPTIONS] RemotePort[:LocalPort] [RemotePort[:LocalPort]]...")
parser.add_option("-t", "--threaded", dest='threaded', action='store_true', default=False,
                  help="use threading to handle multiple connections at once")
parser.add_option("-b", "--bufsize", dest='bufsize', action='store', metavar='KILOBYTES', type='int', default=128,
                  help="specify buffer size for socket forwarding")
parser.add_option("-s", "--socket", dest='sockpath', action='store', metavar='PATH', type='str', default=None,
                  help="specify the path of the usbmuxd socket")

options, args = parser.parse_args()

serverclass = TCPServer
if options.threaded:
    serverclass = ThreadedTCPServer

if len(args) == 0:
    parser.print_help()
    sys.exit(1)

ports = []

for arg in args:
    try:
        if ':' in arg:
            rport, lport = arg.split(":")
            rport = int(rport)
            lport = int(lport)
            ports.append((rport, lport))
        else:
            ports.append((int(arg), int(arg)))
    except:
        parser.print_help()
        sys.exit(1)

servers = []

for rport, lport in ports:
    print("Forwarding local port %d to remote port %d" % (lport, rport))
    server = serverclass((HOST, lport), TCPRelay)
    server.rport = rport
    server.bufsize = options.bufsize
    servers.append(server)

alive = True

while alive:
    try:
        rl, wl, xl = select.select(servers, [], [])
        for server in rl:
            server.handle_request()
    except:
        alive = False



import os
import hashlib
import requests
from .ps_downloader import Ldid
from .ps_logger import Logger


class Hash:
    def get_hash(filePath, url):
        m = hashlib.md5()
        if url == None:
            with open(filePath, 'rb') as fh:
                m = hashlib.md5()
                while True:
                    data = fh.read(8192)
                    if not data:
                        break
                    m.update(data)
                return m.hexdigest()
        else:
            r = requests.get(url)
            for data in r.iter_content(8192):
                m.update(data)
            return m.hexdigest()


class LdidHash:
    def check_linux_64(args, data_dir):
        if args.debug:
            Logger.debug(f"Checking ldid hash...")

        remote_hash = Hash.get_hash(None, Ldid.linux_64_url)
        local_hash = Hash.get_hash(f"{data_dir}/ldid", None)

        if remote_hash == local_hash:
            if args.debug:
                Logger.debug(f"ldid hash successfully verified.")

            return True
        else:
            if args.debug:
                Logger.debug(f"ldid hash failed to verify.")

            return False

    def check_linux_arm64(args, data_dir):
        if args.debug:
            Logger.debug(f"Checking ldid hash...")

        remote_hash = Hash.get_hash(None, Ldid.linux_arm64_url)
        local_hash = Hash.get_hash(f"{data_dir}/ldid", None)

        if remote_hash == local_hash:
            if args.debug:
                Logger.debug(f"ldid hash successfully verified.")

            return True
        else:
            if args.debug:
                Logger.debug(f"ldid hash failed to verify.")

            return False

    def check_macos_64(args, data_dir):
        if args.debug:
            Logger.debug(f"Checking ldid hash...")

        remote_hash = Hash.get_hash(None, Ldid.macos_64_url)
        local_hash = Hash.get_hash(f"{data_dir}/ldid", None)

        if remote_hash == local_hash:
            if args.debug:
                Logger.debug(f"ldid hash successfully verified.")

            return True
        else:
            if args.debug:
                Logger.debug(f"ldid hash failed to verify.")

            return False

    def check_macos_arm64(args, data_dir):
        if args.debug:
            Logger.debug(f"Checking ldid hash...")

        remote_hash = Hash.get_hash(None, Ldid.macos_arm64_url)
        local_hash = Hash.get_hash(f"{data_dir}/ldid", None)

        if remote_hash == local_hash:
            if args.debug:
                Logger.debug(f"ldid hash successfully verified.")

            return True
        else:
            if args.debug:
                Logger.debug(f"ldid hash failed to verify.")

            return False



import pkgutil
from urllib.parse import urlparse


class Copier:
    def copy_postinst(file_path, app_name, in_package):
        """Copy postinst file.

        Args:
            file_path (String): Path of the copy destination.
            app_name (String): Name of the app being processed.
        """

        # Read the file
        if in_package:
            filedata = pkgutil.get_data(__name__, "data/postinst").decode('utf_8')
        else:
            with open("permasigner/data/postinst", 'r') as file:
                filedata = file.read()

        # Replace the target string
        filedata = filedata.replace("{APP_NAME}", app_name)

        # Write the file out again
        with open(file_path, 'w') as file:
            file.write(filedata)

    def copy_postrm(file_path, app_name, in_package):
        """Copy postrm file.

        Args:
            file_path (String): Path of the copy destination.
            app_name (String): Name of the app being processed.
        """

        # Read the file
        if in_package:
            filedata = pkgutil.get_data(__name__, "data/postrm").decode('utf_8')
        else:
            with open('permasigner/data/postrm', 'r') as file:
                filedata = file.read()

        # Replace the target string
        filedata = filedata.replace("{APP_NAME}", app_name)

        # Write the file out again
        with open(file_path, 'w') as file:
            file.write(filedata)

    def copy_control(file_path, app_name, app_bundle, app_version, app_min_ios, app_author, in_package):
        """Copy control file.

        Args:
            file_path (String): Path of the copy destination.
            app_name (String): Name of the app being processed.
            app_bundle (String): Bundle ID of the app being processed.
            app_version (String): Version of the app being processed.
            app_min_ios (String): Minimum iOS version required by the app being processed.
            app_author (String): Author of the app being processed.
        """

        # Read the file
        if in_package:
            filedata = pkgutil.get_data(__name__, "data/control").decode('utf_8')
        else:
            with open('permasigner/data/control', 'r') as file:
                filedata = file.read()

        # Replace the target strings
        filedata = filedata.replace("{APP_NAME}", app_name)
        filedata = filedata.replace("{APP_NAME_ENCODED}", urlparse(app_name).path)
        filedata = filedata.replace("{APP_BUNDLE}", app_bundle)
        filedata = filedata.replace("{APP_VERSION}", app_version)
        filedata = filedata.replace("{APP_MIN_IOS}", app_min_ios)
        filedata = filedata.replace("{APP_AUTHOR}", app_author)

        # Write the file out again
        with open(file_path, 'w') as file:
            file.write(filedata)

    def copy_entitlements(file_path, app_bundle, in_package):
        """Copy entitlements file.

        Args:
            file_path (String): Path of the copy destination.
            app_bundle (String): Bundle ID of the app being processed.
        """

        # Read the file
        if in_package:
            filedata = pkgutil.get_data(__name__, "data/entitlements.plist").decode('utf_8')
        else:
            with open('permasigner/data/entitlements.plist', 'r') as file:
                filedata = file.read()

        # Replace the target string
        filedata = filedata.replace("{APP_BUNDLE}", app_bundle)

        # Write the file out again
        with open(file_path, 'w') as file:
            file.write(filedata)



import subprocess
import time
from getpass import getpass

from paramiko.client import AutoAddPolicy, SSHClient
from paramiko.ssh_exception import AuthenticationException, SSHException, NoValidConnectionsError
from scp import SCPClient
from subprocess import DEVNULL, PIPE

from .ps_logger import Logger, Colors


class Installer:
    def install_deb(args, path_to_deb):
        Logger.log(f'Installing {path_to_deb} to the device', color=Colors.pink)
        print("Relaying TCP connection")
        if args.debug:
            Logger.debug(f"Running command: ./utils/tcprelay.py -t 22:2222")

        relay = subprocess.Popen(
            './utils/tcprelay.py -t 22:2222'.split(), stdout=DEVNULL, stderr=PIPE)
        time.sleep(1)
        try:
            password = getpass(
                prompt="Please provide your user password (default = alpine): ")
            if len(password.strip()) == 0:
                password = 'alpine'
            with SSHClient() as client:
                client.set_missing_host_key_policy(AutoAddPolicy())
                client.connect('localhost',
                               port=2222,
                               username='mobile',
                               password=f'{password}',
                               timeout=5000,
                               allow_agent=False,
                               look_for_keys=False,
                               compress=True)
                with SCPClient(client.get_transport()) as scp:
                    print(f"Sending {path_to_deb} to device")
                    filename = path_to_deb.split("/")[-1]
                    if args.debug:
                        Logger.debug(f"Copying via scp from {path_to_deb} to /var/mobile/Documents/")

                    scp.put(f'{path_to_deb}',
                            remote_path='/var/mobile/Documents')
                    if args.debug:
                        Logger.debug(f"Running command: sudo -nv")

                    stdin, stdout, stderr = client.exec_command('sudo -nv')
                    status = stdout.channel.recv_exit_status()
                    out = stderr.read().decode()

                    if "password" in out:
                        print("User is in sudoers, using sudo")
                        command = f"sudo dpkg -i /var/mobile/Documents/{filename}"
                        if args.debug:
                            Logger.debug(f"Running command: {command}")

                        stdin, stdout, stderr = client.exec_command(
                            f"{command}", get_pty=True)
                        stdin.write(f'{password}\n')
                        stdin.flush()
                        print("Installing... this may take some time")
                        print(stdout.read().decode())

                        if args.debug:
                            Logger.debug(f"Running command: sudo apt-get install -f")
                        # for elucubratus bootstrap
                        streams = client.exec_command(
                            'sudo apt-get install -f', get_pty=True)
                        streams[0].write(f'{password}\n')
                        streams[0].flush()
                        print(streams[1].read().decode())
                    elif status == 0:
                        print('User has nopasswd set')
                        command = f"sudo dpkg -i /var/mobile/Documents/{filename}"
                        if args.debug:
                            Logger.debug(f"Running command: {command}")

                        output = client.exec_command(f'{command}')[1]

                        print("Installing... this may take some time")
                        print(output.read().decode())
                        if args.debug:
                            Logger.debug(f"Running command: sudo apt-get install -f")

                        output = client.exec_command(
                            'sudo apt-get install -f')[1]
                        print(output.read().decode())
                    else:
                        print('Using su command')
                        command = f"su root -c 'dpkg -i /var/mobile/Documents/{filename}'"
                        if args.debug:
                            Logger.debug(f"Running command: {command}")

                        streams = client.exec_command(
                            f"{command}", get_pty=True)

                        output = streams[1].channel.recv(2048).decode()
                        if "password" in output.lower():
                            password = getpass()
                            streams[0].write(f'{password}\n')
                            streams[0].flush()
                            print("Installing... this may take some time")
                            print(streams[1].read().decode())
                            # for elucubratus bootstrap
                            streams = client.exec_command(
                                "su root -c 'apt-get install -f'", get_pty=True)
                            streams[0].write(f'{password}\n')
                            streams[0].flush()
                            print(streams[1].channel.recv(2048).decode())
                        else:
                            print("Installing... this may take some time")
                            print(streams[1].read().decode())
                            if args.debug:
                                Logger.debug(f"Running command: sudo apt-get install -f")

                            output = client.exec_command(
                                'sudo apt-get install -f')[1]
                            print(output.read().decode())
        except (SSHException, NoValidConnectionsError, AuthenticationException) as e:
            Logger.error(e)
        finally:
            relay.kill()



import os
from pathlib import Path
from shutil import copy, copytree
import plistlib
import requests
from urllib.parse import urlparse
import zipfile
import sys
import subprocess
import tempfile
import platform
import argparse
from glob import glob
from subprocess import DEVNULL
import pkgutil

from .ps_copier import Copier
from .ps_hash import LdidHash
from .ps_downloader import DpkgDeb, Ldid
from . import __version__
from .ps_utils import Utils
from .ps_logger import Logger, Colors


if not Utils.is_ios():
    from .ps_usbmux import USBMux
    from .ps_installer import Installer


""" Main Function """


def main(args, in_package=False):
    data_dir = f"{Utils.get_home_data_directory(args)}/.permasigner"
    os.makedirs(data_dir, exist_ok=True)

    if in_package:
        if args.debug:
            Logger.debug(f"Running from package, not cloned repo.")

    ldid_in_path = Utils.cmd_in_path(args, 'ldid')
    dpkg_in_path = Utils.cmd_in_path(args, 'dpkg-deb')
    git_in_path = Utils.cmd_in_path(args, 'git')

    if git_in_path:
        if args.debug:
            Logger.debug(f"Git is in PATH")

        if in_package:
            ver_string = f"{__version__.__version__}"
        elif not "main" in subprocess.getoutput(['git', 'rev-parse', '--abbrev-ref', 'HEAD']):
            ver_string = f"{subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).decode('ascii').strip()}_{subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode('ascii').strip()}"
        else:
            ver_string = f"{__version__.__version__}_rev-{subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode('ascii').strip()}"

    else:
        if args.debug:
            Logger.debug(f"Git is not in PATH")

        ver_string = f"{__version__.__version__}"

    print(Colors.bold + f"Permasigner | Version {ver_string}")
    print(Colors.bold + "Program created by Nebula | Original scripts created by zhuowei | CoreTrust bypass by Linus Henze")
    print()

    # Check if script is running on Windows, if so, fail
    if sys.platform == "windows":
        Logger.error(f"Script must be ran on macOS or Linux.")
        exit(1)

    # Check if codesign is added on Linux or iOS
    if args.codesign:
        if not Utils.is_macos():
            Logger.error(f"Codesign can only be used on macOS, remove the argument to use ldid instead.")
            exit(1)

    # Auto download ldid
    if not ldid_in_path:
        if Path(f"{data_dir}/ldid").exists():
            if Utils.is_linux() and platform.machine() == "x86_64":
                if not LdidHash.check_linux_64(args, data_dir):
                    Logger.log(f"ldid is outdated or malformed, downloading latest version...", color=Colors.pink)
                    os.remove(f"{data_dir}/ldid")
                    Ldid.download_linux_64(args)
            elif Utils.is_linux() and platform.machine() == "aarch64":
                if not LdidHash.check_linux_arm64(args, data_dir):
                    Logger.log(f"ldid is outdated or malformed, downloading latest version...", color=Colors.pink)
                    os.remove(f"{data_dir}/ldid")
                    Ldid.download_linux_arm64(args)
            elif Utils.is_macos() and platform.machine() == "x86_64":
                if not LdidHash.check_macos_64(args, data_dir):
                    Logger.log(f"ldid is outdated or malformed, downloading latest version...", color=Colors.pink)
                    os.remove(f"{data_dir}/ldid")
                    Ldid.download_macos_64(args)
            elif Utils.is_macos() and platform.machine() == "arm64":
                if not LdidHash.check_macos_arm64(args, data_dir):
                    Logger.log(f"ldid is outdated or malformed, downloading latest version...", color=Colors.pink)
                    os.remove(f"{data_dir}/ldid")
                    Ldid.download_macos_arm64(args)
        else:
            Logger.log("ldid binary is not found, downloading latest binary.", color=Colors.pink)
            if Utils.is_linux() and platform.machine() == "x86_64":
                Ldid.download_linux_64(args)
            elif Utils.is_linux() and platform.machine() == "aarch64":
                Ldid.download_linux_arm64(args)
            elif Utils.is_macos() and platform.machine() == "x86_64":
                Ldid.download_macos_64(args)
            elif Utils.is_macos() and platform.machine() == "arm64":
                Ldid.download_macos_arm64(args)

    # Auto download dpkg-deb on Linux
    if not dpkg_in_path and Utils.is_linux():
        if not Path(f"{os.getcwd()}/dpkg-deb").exists():
            if platform.machine() == "x86_64":
                if args.debug:
                    Logger.debug(f"On Linux x86_64, dpkg-deb not found...")

                Logger.log(f"dpkg-deb not found, downloading.", color=Colors.pink)
                DpkgDeb.download_linux_64(args)
                print()
            elif platform.machine() == "aarch64":
                if args.debug:
                    Logger.debug(f"On Linux aarch64, dpkg-deb not found...")

                Logger.log(f"dpkg-deb not found, downloading.", color=Colors.pink)
                DpkgDeb.download_linux_arm64(args)
                print()

    if Utils.is_macos():
        if not subprocess.getstatusoutput("which dpkg")[0] == 0:
            if args.debug:
                Logger.debug(f"On macOS x86_64, dpkg not found...")
            Logger.error("dpkg is not installed and is required on macOS. Install it though brew or Procursus to continue.")
            exit(1)

    # Prompt the user if they'd like to use an external IPA or a local IPA
    if not (args.url or args.path):
        option = Logger.ask("Would you like to use an external or a local IPA? [E, L] ").lower()
        print()

    with tempfile.TemporaryDirectory() as tmpfolder:
        Logger.log(f"Created temporary directory.", color=Colors.pink)
        print()

        # If the user's choice is external, download an IPA
        # Otherwise, copy the IPA to the temporary directory
        if args.url:
            url = args.url

            if not os.path.splitext(urlparse(url).path)[1] == ".ipa":
                Logger.error("URL provided is not an IPA, make sure to provide a direct link.")
                exit(1)

            res = requests.get(url, stream=True)

            try:
                if res.status_code == 200:
                    Logger.log(f"Downloading file...", color=Colors.pink)

                    with open(f"{tmpfolder}/app.ipa", "wb") as f:
                        f.write(res.content)
                else:
                    Logger.error(f"URL provided is not reachable. Status code: {res.status_code}")
                    exit(1)
            except requests.exceptions.RequestException as err:
                Logger.error(f"URL provided is not reachable. Error: {err}")
                exit(1)
        elif args.path:
            path = args.path
            path = path.strip().lstrip("'").rstrip("'")

            if Path(path).exists():
                copy(path, f"{tmpfolder}/app.ipa")
            else:
                priLogger.errornt("That file does not exist! Make sure you're using a direct path to the IPA file.")
                exit(1)
        elif option == "e":
            url = Logger.ask("Paste in the *direct* path to an IPA online: ")

            if not os.path.splitext(urlparse(url).path)[1] == ".ipa":
                Logger.error("URL provided is not an IPA, make sure to provide a direct link.")
                exit(1)

            res = requests.get(url, stream=True)

            try:
                if res.status_code == 200:
                    Logger.log(f"Downloading file...", color=Colors.pink)

                    with open(f"{tmpfolder}/app.ipa", "wb") as f:
                        f.write(res.content)
                else:
                    Logger.error(f"URL provided is not reachable. Status code: {res.status_code}")
                    exit(1)
            except requests.exceptions.RequestException as err:
                Logger.error(f"URL provided is not reachable. Error: {err}")
                exit(1)
        elif option == "l":
            if os.environ.get('IS_DOCKER_CONTAINER', False):
                Logger.info("Running in Docker container, please place an IPA in the 'ipas' folder, then put the name of the file below.")
                ipa_name = input(Colors.orange + '    IPA name (ex. Taurine.ipa, DemoApp.ipa): ' + Colors.reset)
                path = f"/usr/src/permasigner/ipas/{ipa_name}"
            else:
                path = Logger.ask("Paste in the path to an IPA in your file system: ")

            path = path.strip().lstrip("'").rstrip("'")

            if Path(path).exists():
                copy(path, f"{tmpfolder}/app.ipa")
            else:
                Logger.error("That file does not exist! Make sure you're using a direct path to the IPA file.")
                exit(1)
        else:
            Logger.error(f"That is not a valid option!")
            exit(1)
        print()

        # Unzip the IPA file
        Logger.log(f"Unzipping IPA...", color=Colors.pink)
        with zipfile.ZipFile(f"{tmpfolder}/app.ipa", 'r') as f:
            os.makedirs(f"{tmpfolder}/app", exist_ok=False)
            f.extractall(f"{tmpfolder}/app")
        print()

        # Read data from the plist
        Logger.log(f"Reading plist...", color=Colors.pink)

        if Path(f"{tmpfolder}/app/Payload").exists():
            for fname in os.listdir(path=f"{tmpfolder}/app/Payload"):
                if fname.endswith(".app"):
                    app_dir = fname
            print("Found app directory!")
        else:
            Logger.error(f"IPA is not valid!")
            exit(1)

        pre_app_path = os.path.join(f"{tmpfolder}/app/Payload", app_dir)

        if Path(f'{pre_app_path}/Info.plist').exists():
            print("Found Info.plist")
            with open(f'{pre_app_path}/Info.plist', 'rb') as f:
                info = plistlib.load(f)
                if args.name:
                    app_name = args.name
                else:
                    app_name = info["CFBundleName"]

                if args.bundleid:
                    app_bundle = args.bundleid
                else:
                    app_bundle = info["CFBundleIdentifier"]

                app_version = info["CFBundleShortVersionString"]

                if args.minver:
                    app_min_ios = args.minver
                else:
                    app_min_ios = info["MinimumOSVersion"]

                app_author = app_bundle.split(".")[1]
                app_author = app_author[0].upper() + app_author[1:]

                app_executable = info["CFBundleExecutable"]
                print("Found information about the app!")
        print()

        # Get the deb file ready
        Logger.log(f"Preparing deb file...", color=Colors.pink)
        print("Making directories...")
        os.makedirs(f"{tmpfolder}/deb/Applications", exist_ok=False)
        os.makedirs(f"{tmpfolder}/deb/DEBIAN", exist_ok=False)
        print("Copying deb file scripts and control...")
        Copier.copy_postrm(f"{tmpfolder}/deb/DEBIAN/postrm", app_name, in_package)
        Copier.copy_postinst(f"{tmpfolder}/deb/DEBIAN/postinst", app_name, in_package)
        Copier.copy_control(f"{tmpfolder}/deb/DEBIAN/control", app_name,
                            app_bundle, app_version, app_min_ios, app_author, in_package)
        print("Copying app files...")
        full_app_path = os.path.join(f"{tmpfolder}/deb/Applications", app_dir)
        copytree(pre_app_path, full_app_path)
        print("Changing deb file scripts permissions...")
        subprocess.run(
            f"chmod 0755 {tmpfolder}/deb/DEBIAN/postrm".split(), stdout=DEVNULL)
        subprocess.run(
            f"chmod 0755 {tmpfolder}/deb/DEBIAN/postinst".split(), stdout=DEVNULL)
        if app_executable is not None:
            print("Changing app executable permissions...")
            exec_path = os.path.join(full_app_path, app_executable)
            subprocess.run(['chmod', '0755', f'{exec_path}'])
        print()

        # Sign the app
        Logger.log(f"Signing app...", color=Colors.pink)
        Copier.copy_entitlements(f"{tmpfolder}/entitlements.plist", app_bundle, in_package)
        frameworks_path = os.path.join(full_app_path, 'Frameworks')
        if in_package:
            cert_path = Utils.get_resource_path(__name__, "data/certificate.p12")
        else:
            cert_path = "permasigner/data/certificate.p12"

        if args.codesign:
            print("Signing with codesign as it was specified...")
            subprocess.run(
                ['security', 'import', cert_path, '-P', 'password', '-A'], stdout=DEVNULL)

            subprocess.run(['codesign', '-s', 'We Do A Little Trolling iPhone OS Application Signing',
                            '--force', '--deep', '--preserve-metadata=entitlements', f'{full_app_path}'], stdout=DEVNULL)
        else:
            print("Signing with ldid...")
            if Utils.is_ios():
                ldid_cmd = 'ldid'
            else:
                ldid_cmd = f'{data_dir}/ldid'
            if args.debug:
                Logger.debug(
                    f"Running command: {ldid_cmd} -S{tmpfolder}/entitlements.plist -M -K{cert_path} -Upassword '{full_app_path}'")

            subprocess.run([f'{ldid_cmd}', f'-S{tmpfolder}/entitlements.plist', '-M',
                            f'-K{cert_path}', '-Upassword', f'{full_app_path}'], stdout=DEVNULL)

        print()

        # Package the deb file
        Logger.log(f"Packaging the deb file...", color=Colors.pink)
        if args.output:
            path_to_deb = args.output
        elif in_package:
            os.makedirs(f"{Utils.get_home_data_directory(args)}/.permasigner/output", exist_ok=True)
            path_to_deb = f"{Utils.get_home_data_directory(args)}/.permasigner/output/{app_name.replace(' ', '')}.deb"
        else:
            os.makedirs("output", exist_ok=True)
            path_to_deb = f"output/{app_name.replace(' ', '')}.deb"

        dpkg_cmd = f"dpkg-deb -Zxz --root-owner-group -b {tmpfolder}/deb {path_to_deb}"

        if dpkg_in_path:
            if args.debug:
                Logger.debug(f"Path to deb file: {path_to_deb}")
                Logger.debug(f"Running command: {dpkg_cmd}")

            subprocess.run(f"{dpkg_cmd}".split(), stdout=DEVNULL)
        else:
            if args.debug:
                Logger.debug(f"Running command: ./{dpkg_cmd}")

            subprocess.run(f"{data_dir}/{dpkg_cmd}".split(), stdout=DEVNULL)

        is_installed = False
        if not args.noinstall:
            option = 'n'
            if not args.install:
                option = Logger.ask("Would you like install the application to your device (must be connected)? [y, n]: ").lower()

            if option == 'y' or args.install:
                if Utils.is_macos() or Utils.is_linux():
                    try:
                        mux = USBMux()
                        if not mux.devices:
                            mux.process(1.0)
                        if not mux.devices:
                            print("Did not find a connected device")
                        else:
                            print("Found a connected device")
                            Installer.install_deb(args, path_to_deb)
                            is_installed = True
                    except ConnectionRefusedError:
                        print("Did not find a connected device")
                        pass
                elif Utils.is_ios():
                    print("Checking if user is in sudoers")
                    p = subprocess.run('sudo -nv'.split(),
                                       capture_output=True)
                    if p.returncode == 0 or 'password' in p.stderr.decode():
                        print("User is in sudoers, using sudo command")
                        if args.debug:
                            Logger.debug(f"Running command: sudo dpkg -i {path_to_deb}")

                        subprocess.run(
                            ["sudo", "dpkg", "-i", f"{path_to_deb}"], capture_output=True)

                        subprocess.run(
                            ['sudo', 'apt-get', 'install', '-f'], capture_output=True)
                    else:
                        print("User is not in sudoers, using su instead")
                        if args.debug:
                            Logger.debug(f"Running command: su root -c 'dpkg -i {path_to_deb}")

                        subprocess.run(
                            f"su root -c 'dpkg -i {path_to_deb}'".split(), capture_output=True)

                        subprocess.run(
                            "su root -c 'apt-get install -f'".split(), capture_output=True)

    # Done!!!
    print()
    Logger.log(f"We are finished!", color=Colors.green)

    if is_installed:
        Logger.log(f"The application was installed to your device, no further steps are required!", color=Colors.green)
    else:
        Logger.log(f"Copy the newly created deb from the output folder to your jailbroken iDevice and install it!", color=Colors.green)

    Logger.log(f"The app will continue to work when rebooted to stock.", color=Colors.green)
    Logger.log(f"Also, this is free and open source software! Feel free to donate to my Patreon if you enjoy :)", color=Colors.green)
    print(Colors.green + "    https://patreon.com/nebulalol" + Colors.reset)
    Logger.log(f"Output file: {path_to_deb}", color=Colors.green)



class Colors:
    black = '\033[30m'
    red = '\033[31m'
    green = '\033[32m'
    orange = '\033[33m'
    blue = '\033[34m'
    purple = '\033[35m'
    cyan = '\033[36m'
    lightgrey = '\033[37m'
    darkgrey = '\033[90m'
    lightred = '\033[91m'
    lightgreen = '\033[92m'
    yellow = '\033[93m'
    lightblue = '\033[94m'
    pink = '\033[95m'
    lightcyan = '\033[96m'

    reset = '\033[0m'
    bold = '\033[01m'
    disable = '\033[02m'
    underline = '\033[04m'
    reverse = '\033[07m'
    strikethrough = '\033[09m'
    invisible = '\033[08m'


class Logger:
    def log(message, color=None):
        if color is None:
            print(f"[*] {message}")
        else:
            print(color + Colors.bold + "[*] " + Colors.reset + color + f"{message}" + Colors.reset)

    def debug(message):
        print(Colors.lightblue + Colors.bold + "[DEBUG] " + Colors.reset + Colors.lightblue + f"{message}" + Colors.reset)

    def error(message):
        print(Colors.lightred + Colors.bold + "[!] " + Colors.reset + Colors.lightred + f"{message}" + Colors.reset)

    def ask(message):
        return input(Colors.orange + Colors.bold + "[?] " + Colors.reset + Colors.orange + f"{message}" + Colors.reset)



import requests
import subprocess
import os
from shutil import copy, rmtree, move

from .ps_utils import Utils
from .ps_logger import Logger


class DpkgDeb:
    def download_linux_64(args):
        if args.debug:
            Logger.debug(f"Downloading dpkg-deb on Linux x86_64.")

        res = requests.get(
            "http://ftp.us.debian.org/debian/pool/main/d/dpkg/dpkg_1.21.9_amd64.deb", stream=True)
        try:
            if res.status_code == 200:
                with open(f"dpkg.deb", "wb") as f:
                    f.write(res.content)
                    if args.debug:
                        Logger.debug(f"Wrote file.")
            else:
                Logger.error(f"dpkg download URL is not reachable. Status code: {res.status_code}")
                exit(1)
        except requests.exceptions.RequestException as err:
            Logger.error(f"dpkg download URL is not reachable. Error: {err}")
            exit(1)

        subprocess.run(f"ar x dpkg.deb".split(), stdout=subprocess.DEVNULL)
        if args.debug:
            Logger.debug(f"Extracted with ar.")
        subprocess.run(f"tar -xf data.tar.xz".split(),
                       stdout=subprocess.DEVNULL)
        if args.debug:
            Logger.debug(f"Extracted with tar.")
        copy("usr/bin/dpkg-deb", "dpkg-deb")
        if args.debug:
            Logger.debug(f"Copied.")
        subprocess.run(f"chmod +x dpkg-deb".split(), stdout=subprocess.DEVNULL)
        os.remove("data.tar.xz")
        os.remove("control.tar.xz")
        os.remove("debian-binary")
        os.remove("dpkg.deb")
        rmtree("etc")
        rmtree("sbin")
        rmtree("usr")
        rmtree("var")
        move("dpkg-deb", f"{Utils.get_home_data_directory(args)}/.permasigner/dpkg-deb")
        if args.debug:
            Logger.debug(f"Cleaned up.")

    def download_linux_arm64(args):
        if args.debug:
            Logger.debug(f"Downloading dpkg-deb on Linux aarch64.")

        res = requests.get(
            "http://ftp.us.debian.org/debian/pool/main/d/dpkg/dpkg_1.21.9_arm64.deb", stream=True)
        try:
            if res.status_code == 200:
                with open(f"dpkg.deb", "wb") as f:
                    f.write(res.content)
                    if args.debug:
                        Logger.debug(f"Wrote file.")
            else:
                Logger.error(f"dpkg download URL is not reachable. Status code: {res.status_code}")
                exit(1)
        except requests.exceptions.RequestException as err:
            Logger.error(f"dpkg download URL is not reachable. Error: {err}")
            exit(1)

        subprocess.run(f"ar x dpkg.deb".split(), stdout=subprocess.DEVNULL)
        if args.debug:
            Logger.debug(f"Extracted with ar.")
        subprocess.run(f"tar -xf data.tar.xz".split(),
                       stdout=subprocess.DEVNULL)
        if args.debug:
            Logger.debug(f"Extracted with tar.")
        copy("usr/bin/dpkg-deb", "dpkg-deb")
        if args.debug:
            Logger.debug(f"Copied.")
        subprocess.run(f"chmod +x dpkg-deb".split(), stdout=subprocess.DEVNULL)
        os.remove("data.tar.xz")
        os.remove("control.tar.xz")
        os.remove("debian-binary")
        os.remove("dpkg.deb")
        rmtree("etc")
        rmtree("sbin")
        rmtree("usr")
        rmtree("var")
        move("dpkg-deb", f"{Utils.get_home_data_directory(args)}/.permasigner/dpkg-deb")
        if args.debug:
            Logger.debug(f"Cleaned up.")


class Ldid:
    linux_64_url = "https://github.com/ProcursusTeam/ldid/releases/latest/download/ldid_linux_x86_64"
    linux_arm64_url = "https://github.com/ProcursusTeam/ldid/releases/latest/download/ldid_linux_aarch64"
    macos_64_url = "https://github.com/ProcursusTeam/ldid/releases/latest/download/ldid_macos_x86_64"
    macos_arm64_url = "https://github.com/ProcursusTeam/ldid/releases/latest/download/ldid_macos_arm64"

    def download_linux_64(args):
        if args.debug:
            Logger.debug(f"Downloading ldid on Linux x86_64.")

        res = requests.get(Ldid.linux_64_url, stream=True)
        try:
            if res.status_code == 200:
                with open(f"ldid", "wb") as f:
                    f.write(res.content)
                    if args.debug:
                        Logger.debug(f"Wrote file.")
            else:
                Logger.error(f"ldid download URL is not reachable. Status code: {res.status_code}")
                exit(1)
        except requests.exceptions.RequestException as err:
            Logger.error(f"ldid download URL is not reachable. Error: {err}")
            exit(1)

        subprocess.run(f"chmod +x ldid".split(), stdout=subprocess.DEVNULL)
        move("ldid", f"{Utils.get_home_data_directory(args)}/.permasigner/ldid")

    def download_linux_arm64(args):
        if args.debug:
            Logger.debug(f"Downloading ldid on Linux aarch64.")

        res = requests.get(Ldid.linux_arm64_url, stream=True)
        try:
            if res.status_code == 200:
                with open(f"ldid", "wb") as f:
                    f.write(res.content)
                    if args.debug:
                        Logger.debug(f"Wrote file.")
            else:
                Logger.error(f"ldid download URL is not reachable. Status code: {res.status_code}")
                exit(1)
        except requests.exceptions.RequestException as err:
            Logger.error(f"ldid download URL is not reachable. Error: {err}")
            exit(1)

        subprocess.run(f"chmod +x ldid".split(), stdout=subprocess.DEVNULL)
        move("ldid", f"{Utils.get_home_data_directory(args)}/.permasigner/ldid")

    def download_macos_64(args):
        if args.debug:
            Logger.debug(f"Downloading ldid on macOS x86_64.")

        res = requests.get(Ldid.macos_64_url, stream=True)
        try:
            if res.status_code == 200:
                with open(f"ldid", "wb") as f:
                    f.write(res.content)
                    if args.debug:
                        Logger.debug(f"Wrote file.")
            else:
                Logger.error(f"ldid download URL is not reachable. Status code: {res.status_code}")
                exit(1)
        except requests.exceptions.RequestException as err:
            Logger.error(f"ldid download URL is not reachable. Error: {err}")
            exit(1)

        subprocess.run(f"chmod +x ldid".split(), stdout=subprocess.DEVNULL)
        move("ldid", f"{Utils.get_home_data_directory(args)}/.permasigner/ldid")

    def download_macos_arm64(args):
        if args.debug:
            Logger.debug(f"Downloading ldid on macOS arm64.")

        res = requests.get(Ldid.macos_arm64_url, stream=True)
        try:
            if res.status_code == 200:
                with open(f"ldid", "wb") as f:
                    f.write(res.content)
                    if args.debug:
                        Logger.debug(f"Wrote file.")
            else:
                Logger.error(f"ldid download URL is not reachable. Status code: {res.status_code}")
                exit(1)
        except requests.exceptions.RequestException as err:
            Logger.error(f"ldid download URL is not reachable. Error: {err}")
            exit(1)

        subprocess.run(f"chmod +x ldid".split(), stdout=subprocess.DEVNULL)
        move("ldid", f"{Utils.get_home_data_directory(args)}/.permasigner/ldid")



from os.path import join
import setuptools

with open("README.md", "r") as fh:
    readme = fh.read()

exec(open('permasigner/__version__.py').read())


def parse_requirements(requirements, ignore=('setuptools',)):
    with open(requirements) as f:
        packages = set()
        for line in f:
            line = line.strip()
            if line.startswith(('#', '-r', '--')):
                continue
            if '#egg=' in line:
                line = line.split('#egg=')[1]
            pkg = line.strip()
            if pkg not in ignore:
                packages.add(pkg)
        return tuple(packages)


setuptools.setup(
    name='permasigner',
    version=__version__,
    description=('Permanently signs IPAs on jailbroken iDevices (persists on stock).'),
    license='BSD-3-Clause',
    url='https://github.com/itsnebulalol/permasigner',
    python_requires=">=3.7",
    packages=setuptools.find_packages(),
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Environment :: Console',
        'Operating System :: OS Independent',
    ],
    scripts=['bin/permasigner'],
    keywords='python, windows, macos, linux, docker, cli, open-source, ios, command-line-app, cli-app, hacktoberfest, procursus, permasign, permasigner',
    include_package_data=True,
    author='Nebula',
    install_requires=parse_requirements('requirements.txt'),
    author_email='me@itsnebula.net',
    zip_safe=False,
)



import sys
import subprocess
import platform
import os
import importlib

from .ps_logger import Logger


class Utils:
    def is_ios():
        if not sys.platform == "darwin":
            return False

        return platform.machine().startswith("i")

    def cmd_in_path(args, cmd):
        if args.debug:
            Logger.debug(f"Checking if command {cmd} is in PATH...")

        if cmd == "ldid":
            if Utils.is_ios():
                if args.debug:
                    Logger.debug(f"Checking for ldid on iOS")

                if os.path.exists("/.bootstrapped"):
                    Logger.error("Your device seems to be strapped with Elucubratus. Unfortunately, we do not support these devices. You can switch to a device that uses Procursus (Taurine, odysseyra1n), or use the online method on our GitHub.")
                    print("    https://github.com/itsnebulalol/permasigner/wiki/Run-Online")
                    exit(1)

                if Utils.is_dpkg_installed("ldid"):
                    if args.debug:
                        Logger.debug(f"ldid is installed via dpkg")

                    return True
                else:
                    Logger.error("[-] ldid is required on iOS, but it is not installed. Please install it from Procursus.")
                    exit(1)

            # It seems like a better idea to force download ldid on macOS and Linux to make sure
            # they have the proper version all the time. A lot of ugly code is necessary for it.
            return False

        return subprocess.getstatusoutput(f"which {cmd}")[0] == 0

    def is_macos():
        if platform.machine().startswith("i"):
            return False

        return sys.platform == "darwin"

    def is_linux():
        return sys.platform == "linux"

    def is_dpkg_installed(pkg):
        return (os.system("dpkg -s " + pkg + "> /dev/null 2>&1")) == 0

    def get_home_data_directory(args):
        if os.environ.get("XDG_DATA_HOME"):
            if args.debug:
                Logger.debug(f"Using XDG_DATA_HOME: {os.environ.get('XDG_DATA_HOME')}")
            return os.environ.get("XDG_DATA_HOME")
        else:
            if args.debug:
                Logger.debug(f"Using user home directory: {os.path.expanduser('~')}")
            return os.path.expanduser('~')

    def get_resource_path(package, resource):
        spec = importlib.util.find_spec(package)
        if spec is None:
            return None
        loader = spec.loader
        if loader is None or not hasattr(loader, 'get_data'):
            return None
        mod = (sys.modules.get(package) or
               importlib._bootstrap._load(spec))
        if mod is None or not hasattr(mod, '__file__'):
            return None

        parts = resource.split('/')
        parts.insert(0, os.path.dirname(mod.__file__))
        return f"{os.path.join(*parts)}"
